"""MasterTherm Connection to the Web API."""
from datetime import datetime
from hashlib import sha1
import logging
from urllib.parse import urljoin
from json.decoder import JSONDecodeError

from masterthermconnect.const import (
    APP_CLIENTINFO,
    APP_OS,
    APP_VERSION,
    COOKIE_TOKEN,
    DATE_FORMAT,
    URL_BASE,
    URL_LOGIN,
    URL_PUMPDATA,
    URL_PUMPINFO,
)

from .exceptions import (
    MasterThermAuthenticationError,
    MasterThermConnectionError,
    MasterThermResponseFormatError,
    MasterThermTokenInvalid,
)

_LOGGER = logging.getLogger(__name__)

class Connection:
    """Connection Handler for the MasterTherm API."""

    def __init__(self, websession, username, password):
        """Initiate the Connection API."""
        self.__session = websession
        self.__unname = username
        self.__upwd = sha1(password.encode("utf-8")).hexdigest()
        self.__clientinfo = APP_CLIENTINFO
        self.__token = None
        self.__expires = None
        self.__is_connected = False

    async def __post(self, url, params):
        """Perform the Post to the API server and check results."""
        # Check to see if our token is still valid, if not re-connect
        if self.__expires is None:
            await self.connect()
        if datetime.now() > self.__expires:
            await self.connect()

        # Post with the valid token
        cookies = {COOKIE_TOKEN: self.__token, "$version": APP_VERSION}
        try:
            response = await self.__session.post(
                urljoin(URL_BASE,url),
                data=params,
                headers={"content-type": "application/x-www-form-urlencoded"},
                cookies=cookies
            )
        except Exception as ex:
            _LOGGER.error("MasterTherm API some other error: %s", ex)
            raise Exception() from ex

        if response.status != 200:
            error_msg = await response.text()
            raise MasterThermConnectionError(str(response.status),error_msg)

        # Convert to JSON if possible and return the result
        try:
            response_json = await response.json()
        except JSONDecodeError as exc:
            response_text = await response.text()
            if response_text == "User not logged in":
                _LOGGER.error("MasterTherm API Invalid Token: %s", response_text)
                raise MasterThermTokenInvalid("1", response_text) from exc
            else:
                _LOGGER.error("MasterTherm API some other error: %s", response_text)
                raise MasterThermResponseFormatError("2",response_text) from exc

        return response_json

    async def connect(self):
        """Perform the connection to the Mastertherm API Server:

        Returns:
             json_repsonse (dict): Devices and Information in JSON Format

        Raises:
            MasterThermConnectionError - Failed to Connect
            MasterThermAuthenticationError - Failed to Authenticate"""

        # TODO Check Reconnection on Expire
        self.__is_connected = False

        params = f"login=login&uname={self.__unname}&upwd={self.__upwd}&{self.__clientinfo}"
        response = await self.__session.post(
            urljoin(URL_BASE,URL_LOGIN),
            data=params,
            headers={"content-type": "application/x-www-form-urlencoded"}
        )

        # Response should always be 200 even for login failures.
        if response.status != 200:
            error_msg = await response.text()
            raise MasterThermConnectionError(str(response.status), error_msg)

        # Expect that the response is JSON, check the result.
        response_json = await response.json()
        if response_json["returncode"] != 0:
            raise MasterThermAuthenticationError(
                response_json["returncode"],response_json["message"])

        # Get or Refresh the Token and Expiry
        cookie = response.cookies.get(COOKIE_TOKEN)
        self.__token = cookie.value
        self.__expires = datetime.strptime(cookie["expires"], DATE_FORMAT)

        self.__is_connected = True
        return response_json

    async def is_connected(self):
        """Check we are connected to the API and still valid."""
        # Check expiry, maybe just get info to check.
        return self.__is_connected

    async def get_device_info(self, module_id, device_id):
        """Return the Pump Information as sent from server."""
        params = f"moduleid={module_id}&unitid={device_id}&application={APP_OS}"
        response = await self.__post(URL_PUMPINFO,params)
        return response

    async def get_device_data(self, module_id, device_id, last_update_time = None):
        """Return the Device Data from the server."""
        params = f"moduleId={module_id}&deviceId={device_id}&application={APP_OS}&"
        if last_update_time is None:
            params = params + "messageId=1&lastUpdateTime=0&errorResponse=true&fullRange=true"
        else:
            params = params + \
                f"messageId=2&lastUpdateTime={last_update_time}&errorResponse=true&fullRange=true"
        response = await self.__post(URL_PUMPDATA,params)
        return response



"""Contains all the MasterTherm Exception classes."""

class MasterThermError(Exception):
    """Base Exception for all MasterTherm Exceptions"""
    def __init__(self,status, message):
        """Initialize."""
        super().__init__(status)
        self.status = status
        self.message = message

class MasterThermAuthenticationError(MasterThermError):
    """Raised when login returns wrong result."""

class MasterThermConnectionError(MasterThermError):
    """Raised when communication ended in error."""

class MasterThermResponseFormatError(MasterThermError):
    """Raised when page returns an unexpected response."""

class MasterThermTokenInvalid(MasterThermError):
    """Raised when page returns the Token is not valie."""

class MasterThermUnsupportedRole(MasterThermError):
    """Raised when connecting to an Unsupported Role."""



"""Main Program, for Testing Mainly"""
import argparse
import asyncio
import getpass

from aiohttp import ClientSession

from masterthermconnect.__version__ import __version__
from masterthermconnect.controller import Controller
from masterthermconnect.exceptions import MasterThermError

def get_arguments() -> argparse.Namespace:
    """Read the Arguments passed in."""
    parser = argparse.ArgumentParser(
        prog="masterthermconnect",
        description="Python MasterTherm Connect API Module.",
        formatter_class=argparse.MetavarTypeHelpFormatter
    )
    parser.add_argument(
        "--version", action="version",
        version="MasterTherm Connect API Version: "+__version__,
        help="display the MasterTherm Connect API version"
    )
    parser.add_argument(
        "--user", type=str,
        help="login user for MasterTherm"
    )
    parser.add_argument(
        "--password", type=str, default=None,
        help="login password for MasterTherm"
    )
    parser.add_argument(
        "--list-devices", action="store_true",
        help="list the devices connected to the account"
    )

    arguments = parser.parse_args()
    return arguments

async def list_devices(user, password) -> int:
    """List Devices after Connection"""
    session = ClientSession()
    controller = Controller(session, user, password)

    try:
        await controller.connect()
    except MasterThermError as mte:
        print("Connection Failed " + mte.message)
        return mte.status

    print(controller.get_devices())

    await session.close()
    return 0


def main() -> int:
    """Allow for some testing of connections from Command Line."""
    login_user = ""
    login_pass = ""
    args = get_arguments()

    if args.user is not None:
        login_user = args.user
    else:
        login_user = input("User: ")

    if args.password is not None:
        login_pass = args.password
    else:
        login_pass = getpass.getpass()

    if args.list_devices:
        asyncio.run(list_devices(login_user, login_pass))

    return 0

if __name__ == "__main__":
    main()



"""Constants used by the module."""
APP_CLIENTINFO = "os=android&osversion=7.0&ver=8&info=Google%2CAndroid"
APP_OS = "android"
APP_VERSION = "1"

COOKIE_TOKEN = "PHPSESSID"
DATE_FORMAT = "%a, %d-%b-%Y %H:%M:%S %Z"
SUPPORTED_ROLES = ["400"]

URL_BASE = "https://mastertherm.vip-it.cz"
URL_LOGIN = "/plugins/mastertherm_login/client_login.php"
URL_PUMPINFO = "/plugins/get_pumpinfo/get_pumpinfo.php"
URL_PUMPDATA = "/mt/PassiveVizualizationServlet"

# Used to setup the pad name, for some reason they don't like the letter Q
CHAR_MAP = ["-","A","B","C","D","E","F","G",
    "H","I","J","K","L","M","N","O","P","R",
    "S","T","U","V","W","X","Y","Z"]

DEVICE_SWITCH_MAP = {
    0: "D_348",
    1: "D_433",
    2: "D_326",
    3: "D_316",
    4: "D_307",
    5: "D_298",
    6: "D_436",
    7: "D_278",
    8: "D_182",
}

PAD_MAP = {
    0: "heating",
    1: "cooling",
    2: "padf",
    3: "pade",
    4: "padd",
    5: "padc",
    6: "padb",
    7: "pada",
    8: "padz",
}

DEVICE_INFO_MAP = {
    "name": "givenname",
    "surname": "surname",
    "country": "localization",
    "language": "lang",
    "hp_type": "type",
    "controller": "regulation",
    "exp": "exp",
    "output": "output",
    "reservation": "reservation",
    "place": "city",
    "latitude": "password9",
    "longitude": "password10",
    "notes": "notes",
    "pada": "pada",
    "padb": "padb",
    "padc": "padc",
    "padd": "padd",
    "pade": "pade",
    "padf": "padf",
    "padz": "padz",
}

DEVICE_DATA_PADMAP = {
    "pada": {
        "enabled": ["fixed",False],
        "name": ["string",["I_211", "I_212", "I_213", "I_214", "I_215", "I_216"]],
        "on": ["bool","D_212"],
        "pump_running": ["bool",""],
        "water_temp": ["float","A_126"],
        "water_requested": ["decimal",""],
        "ambient_temp": ["float",""],
        "ambient_requested": ["float",""],
        "control_curve": {
            "setpoint_a_outside": ["float","A_101"],
            "setpoint_a_requested": ["float","A_106"],
            "setpoint_b_outside": ["float","A_102"],
            "setpoint_b_requested": ["float","A_107"],
        },
    },
    "padb": {
        "enabled": ["fixed",False],
        "name": ["string",["I_221", "I_222", "I_223", "I_224", "I_225", "I_226"]],
        "on": ["bool","D_216"],
        "pump_running": ["bool",""],
        "water_temp": ["float","A_91"],
        "water_requested": ["decimal",""],
        "ambient_temp": ["float",""],
        "ambient_requested": ["float",""],
        "control_curve": {
            "setpoint_a_outside": ["float","A_108"],
            "setpoint_a_requested": ["float","A_84"],
            "setpoint_b_outside": ["float","A_109"],
            "setpoint_b_requested": ["float","A_85"],
        },
    },
    "padc": {
        "enabled": ["fixed",False],
        "name": ["string",["I_231", "I_232", "I_233", "I_234", "I_235", "I_236"]],
    },
    "padd": {
        "enabled": ["fixed",False],
        "name": ["string",["I_241", "I_242", "I_243", "I_244", "I_245", "I_246"]],
    },
    "pade": {
        "enabled": ["fixed",False],
        "name": ["string",["I_251", "I_252", "I_253", "I_254", "I_255", "I_256"]],
    },
    "padf": {
        "enabled": ["fixed",False],
        "name": ["string",["I_261", "I_262", "I_263", "I_264", "I_265", "I_266"]],
    },
}

# NOTES --------------------------------------------------
# pswitch_data_1: Cooling Curve
# a_outside: A_47
# b_outside: A_48
# a_requested: A_49
# b_requested: A_50
# --------------------------------------------------------
DEVICE_DATA_MAP = {
    "on": ["bool","D_3"],
    "heating": ["bool",""],
    "compressor_running": ["bool",""],
    "circulation_pump_running": ["bool",""],
    "defrost_mode": ["bool",""],
    "outside_temp": ["float","A_3"],
    "requested_temp": ["float","A_1"],
    "actual_temp": ["float","A_90"],
    "compressor_run_time": ["int","I_11"],
    "compressor_start_counter": ["int","I_12"],
    "pump_runtime": ["int","I_13"],
    "heating_curve": {
        "setpoint_a_outside": ["float","A_35"],
        "setpoint_a_requested": ["float","A_37"],
        "setpoint_b_outside": ["float","A_36"],
        "setpoint_b_requested": ["float","A_38"],
    },
    "pads": DEVICE_DATA_PADMAP,
}



"""MasterTherm Controller, for handling MasterTherm Data."""
import logging

from .connection import Connection
from .const import (
    CHAR_MAP,
    DEVICE_DATA_MAP,
    DEVICE_DATA_PADMAP,
    DEVICE_INFO_MAP,
    DEVICE_SWITCH_MAP,
    PAD_MAP,
    SUPPORTED_ROLES,
)
from .exceptions import MasterThermUnsupportedRole

_LOGGER = logging.getLogger(__name__)

class Controller:
    """Mastertherm Connect Object."""

    def __init__(self, websession, username, password):
        """Initialize the Connection Object."""
        self.__api = Connection(websession, username, password)
        self.__device_map = None
        self.__inverted_map = None
        self.__role = ""
        self.__data_loaded = False

        # The device structure is held as a dictionary with the following format:
        # {
        #   "id-mb_addr": {
        #       "lastUpdateTime": "1234567890",
        #       "info": { Various Information },
        #       "data": { Normalized Data Information }
        #       "updatedData": { All Updated Data since last update },
        #       "fullData": { Full Data including last updated },
        #   }
        # }
        self.__devices = {}

    def __invert_device_map(self, device_map, key_list = None):
        """Invert the given map and return, this is a nested method."""

        if key_list is None:
            key_list = []

        inverted_map = {}
        for key, item in device_map.items():
            new_key_list = key_list.copy()
            new_key_list.append(key)
            if not isinstance(item,dict):
                item_value = item[1]
                if isinstance(item_value,list):
                    for list_value in item_value:
                        inverted_map[list_value] = new_key_list
                else:
                    item_type = item[0]
                    if not (item_value == "" or item_type == "fixed"):
                        inverted_map[item_value] = new_key_list
            else:
                inverted_map.update(self.__invert_device_map(item, new_key_list))

        return inverted_map

    def __populate_data(self, device_map, registers):
        """Populate the Data from the fullData and DeviceMap."""
        data = {}
        for key, item in device_map.items():
            if not isinstance(item, dict):
                item_type = item[0]
                item_value = item[1]
                if item_type == "fixed":
                    data[key] = item_value
                elif item_type == "bool":
                    if item_value == "":
                        data[key] = False
                    else:
                        data[key] = (registers[item_value] == "1")
                elif item_type == "float":
                    if item_value == "":
                        data[key] = 0.0
                    else:
                        data[key] = float(registers[item_value])
                elif item_type == "int":
                    if item_value == "":
                        data[key] = 0
                    else:
                        data[key] = int(registers[item_value])
                elif item_type == "string":
                    if item_value == "":
                        data[key] = ""
                    else:
                        item_str = ""
                        for list_value in item_value:
                            item_str = item_str + CHAR_MAP[int(registers[list_value])]
                        data[key] = item_str
            else:
                data[key] = self.__populate_data(device_map[key], registers)

        return data

    def __get_pad_name(self, pad, device_key):
        """Build the Pad Name from the full data."""
        if pad not in DEVICE_DATA_PADMAP:
            return "0"

        pad_name = ""
        pad_empty = ""
        full_data = self.__devices[device_key]["fullData"]
        for key in DEVICE_DATA_PADMAP[pad]["name"][1]:
            pad_name = pad_name + CHAR_MAP[int(full_data[key])]
            pad_empty = pad_empty + "-"

        if pad_name == pad_empty:
            pad_name = "0"
        return pad_name

    def __enabled_pads(self, device_key):
        """Enable the Pads for the devices, decoded as best as possible."""
        full_data = self.__devices[device_key]["fullData"]
        pad_info = {}

        # Pad 0 and 1 I believe are the Main Heating and Cooling circuites
        # I have heating enabled and cooling disabled but both show disabled.
        # TODO: Check what happens if I enable/ Disable what registers get updated.
        for i in range(2):
            pad_info[PAD_MAP[i]] = (full_data[DEVICE_SWITCH_MAP[i]] == "1")

        # Following Code is to manage Switch Pad a to f which
        # these are various typs of circuites Water Heating/ Thermal etc.
        # TODO: see if we can identify where the types are stored.

        # Used in the process of Enabling/ Disabling Pads.
        check_code = int(full_data["I_104"])
        if check_code < 11:
            check_code = 10
        else:
            if check_code <= 200:
                check_code = 11

        # Setup and Enable Pad a to f
        for i in range(2,8):
            pad_name = self.__get_pad_name(PAD_MAP[i], device_key)
            if pad_name != "0" and check_code >= 11:
                pad_info[PAD_MAP[i]] = (full_data[DEVICE_SWITCH_MAP[i]] == "1")
            else:
                pad_info[PAD_MAP[i]] = False

        # Pad Switch 8, no idea where a name comes from so its always disabled.
        # maby its SOLAR?
        pad_name = self.__get_pad_name(PAD_MAP[8], device_key)
        pad_info[PAD_MAP[8]] = False
        if pad_name != "0":
            if check_code >= 11:
                # If any switch a to f is enabled then enable.
                for i in range(7,1,-1):
                    if pad_info[PAD_MAP[i]]:
                        pad_info[PAD_MAP[8]] = True
                        break
            else:
                pad_info[PAD_MAP[8]] = (
                    full_data[DEVICE_SWITCH_MAP[8]] == "1" and
                    float(full_data["A_190"]) > 0.1
                )

        return pad_info

    async def __full_load(self):
        """Perform a full load and create structure."""
        self.__data_loaded = False

        for device_key, device in self.__devices.items():
            module_id = device["info"]["module_id"]
            device_id = device["info"]["device_id"]

            # Create the Device Info
            device_info = await self.__api.get_device_info(module_id, device_id)
            for key, item in DEVICE_INFO_MAP.items():
                if item in device_info:
                    device["info"][key] = device_info[item]

            # Get the Full Device Data
            device_data = await self.__api.get_device_data(module_id, device_id)
            device["lastUpdateTime"] = device_data["timestamp"]
            device["updatedData"] = device_data["data"]["varfile_mt1_config1"]["001"].copy()
            device["fullData"] = device["updatedData"].copy()

            # Construct Normalized Data, using device map.
            self.__device_map = DEVICE_DATA_MAP
            enabled_pads = self.__enabled_pads(device_key)
            for pad, pad_enabled in enabled_pads.items():
                if not pad_enabled:
                    self.__device_map["pads"].pop(pad, None)

            self.__inverted_map = self.__invert_device_map(self.__device_map)
            device["data"] = self.__populate_data(self.__device_map, device["fullData"])

        self.__data_loaded = True
        return True

    async def connect(self, update_data = True):
        """Connect to the API, check the supported roles and update if required.

        Parameters:
            update_data (bool): False to only connect, default True

        Returns:
            connected (bool): True if connected Raises Error if not

        Raises:
            MasterThermConnectionError - Failed to Connect
            MasterThermAuthenticationError - Failed to Authenticate
            MasterThermUnsportedRole - Role is not supported by API"""
        result = await self.__api.connect()
        self.__role = result["role"]

        if self.__role not in SUPPORTED_ROLES:
            raise MasterThermUnsupportedRole("2", "Unsupported Role " + result["role"])

        # Initialize the Dictionary.
        self.__devices = {}
        for module in result["modules"]:
            for device in module["config"]:
                device_key = module["id"] + "-" + device["mb_addr"]

                self.__devices[device_key] = {
                    "lastUpdateTime": "0",
                    "info": {
                        "module_id": module["id"],
                        "module_name": module["module_name"],
                        "device_id": device["mb_addr"],
                        "device_name": device["mb_name"],
                    },
                    "updatedData": {},
                    "fullData": {},
                    "data": {},
                }

        if update_data:
            return await self.__full_load()

        return True

    async def refresh(self,full_load = False):
        """Refresh or Reload all entries for all devices."""
        if not await self.__api.is_connected():
            return False

        if full_load:
            return self.__full_load()

        if not self.__data_loaded:
            return False

        for device in self.__devices.values():
            module_id = device["info"]["module_id"]
            device_id = device["info"]["device_id"]

            # Refresh Device Info (checks login too)
            device_info = await self.__api.get_device_info(module_id, device_id)
            if device_info["returncode"] == "0":
                for key, item in DEVICE_INFO_MAP.items():
                    if item in device_info:
                        device["info"][key] = device_info[item]

            device_data = await self.__api.get_device_data(
                module_id, device_id, last_update_time=device["lastUpdateTime"])

            device["lastUpdateTime"] = device_data["timestamp"]
            device["updatedData"] = device_data["data"]["varfile_mt1_config1"]["001"].copy()
            device["fullData"].update(device["updatedData"])

            # Refresh Normalized Data
            update_data = False
            for register_key in device["updatedData"]:
                if register_key in self.__inverted_map:
                    update_data = True
                    break

            if update_data:
                device["data"] = self.__populate_data(self.__device_map, device["fullData"])

        return True

    def get_devices(self):
        """Return a List of the Devices with plus information."""
        # TODO: Add Randomize Key Data
        device_return = {}
        for device_id, device in self.__devices.items():
            device_return[device_id] = device["info"]

        return device_return

    def get_device_info(self, module_id, device_id):
        """Get the Information for a specific device."""
        # TODO: Add Randomize Key Data
        info = {}
        key = module_id + "-" + device_id
        if key in self.__devices:
            info = self.__devices[key]["info"]
        return info

    def get_device_registers(self, module_id, device_id, last_updated = False):
        """Get the Device Register Data, if lastUpdated is True then get the latest update data."""
        data = {}
        key = module_id + "-" + device_id
        if key in self.__devices:
            if last_updated:
                data = self.__devices[key]["updatedData"]
            else:
                data = self.__devices[key]["fullData"]
        return data

    def get_device_data(self, module_id, device_id):
        """Get the Device Data, if lastUpdated is True then get the latest update data."""
        data = {}
        key = module_id + "-" + device_id
        if key in self.__devices:
            data = self.__devices[key]["data"]
        return data



"""Python API wrapper for Mastertherm Connect."""
from masterthermconnect.__version__ import __version__
from masterthermconnect.connection import Connection
from masterthermconnect.controller import Controller
from masterthermconnect.exceptions import (
    MasterThermAuthenticationError,
    MasterThermConnectionError,
    MasterThermResponseFormatError,
    MasterThermTokenInvalid,
    MasterThermUnsupportedRole,
)

__all__ = [
    "__version__",
    "Connection",
    "Controller",
    "MasterThermAuthenticationError",
    "MasterThermConnectionError",
    "MasterThermResponseFormatError",
    "MasterThermTokenInvalid",
    "MasterThermUnsupportedRole",
]
